//===------------------------ParseStmtOpenMP.cpp ---------------------------===/
//              /{ OpenMP pragma directives statement Parser  }/
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
// This file implements the OpenMP pragma directive statements parsing portions
// of the Parser interface.
//
//===----------------------------------------------------------------------===//

#include "clang/Parse/Parser.h"
#include "clang/Sema/DeclSpec.h"
#include "clang/Sema/PrettyDeclStackTrace.h"
#include "clang/Sema/Scope.h"
#include "clang/Sema/TypoCorrection.h"
#include "clang/Basic/Diagnostic.h"
#include "clang/Basic/PrettyStackTrace.h"
#include "clang/Basic/SourceManager.h"
#include "llvm/ADT/SmallString.h"
using namespace clang;

//===----------------------------------------------------------------------===//
// OpenMP 3.x: OpenMP pragma directive statements. 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse \#pragma omp parallel ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpParallel(SourceLocation *TrailingElseLoc) {
  assert(Tok.is(tok::pragma_omp_parallel) && 
                "Not a pragma omp \"parallel\" directive statement");

  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return StmtEmpty();

  // FIXME: Do not enable it while submitting firt patch.
/*
  // It could be /{ \#pragma omp parallel for ... }/. Look ahead one token, and
  // if it is a kewword "for", then parse /{ \#pragma omp parallel for ... }/.
  if (NextToken().is(tok::kw_for))
  {
    // Parse \#pragam omp parallel for ...
    return ParsePragmaOmpParallelFor(TrailingElseLoc);
  }

  // Record the source location information for the word "parallel".
  SourceLocation ParallelLoc = Tok.getLocation();
  
  // Eat-up the word "parallel".
  ConsumeToken(); 
  
  // FIXME: Scope??

  // Parse all the OpenMP clasues in the current OpenMP pragma directive line.
  StmtResult Clauses(ParsePragmaOmpClauses());

  // FIXME: Any other missing checks?

  // Parse the parallel body.
  // FIXME: Is it correct??
  StmtResult ParallelBody(ParseStatement(TrailingElseLoc));

  // FIXME: Errors??

  //FIXME - Based on the changes within this routine.
  return Actions.ActOnPragmaOmpParallel(ParallelLoc, Clauses.get(),
                                        ParallelBody.get());
*/
}

//===----------------------------------------------------------------------===//
// Parse \#pragam omp for ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpFor(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_for) && 
                "Not a pragma omp \"for\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpFor();
}

//===----------------------------------------------------------------------===//
// Parse \#pragam omp parallel for ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpParallelFor(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_parallel) &&
                "Not a pragma omp \"parallel for\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpParallelFor();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp sections ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpSections(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_sections) && 
                "Not a pragma omp \"sections\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpSections();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp section
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpSection(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_section) &&
                "Not a pragma omp \"section\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpSection();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp single ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpSingle(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_single) &&
                "Not a pragma omp \"single\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpSingle(); 
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp task ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpTask(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_task) &&
                "Not a pragma omp \"task\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpTask();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp master
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpMaster(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_master) &&
                "Not a pragma omp \"master\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpMaster();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp critical ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpCritical(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_critical) &&
                "Not a pragma omp \"critical\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpCritical();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp barrier
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpBarrier(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_barrier) &&
                "Not a pragma omp \"barrier\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpBarrier();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp taskwait
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpTaskwait(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_taskwait) &&
                "Not a pragma omp \"taskwait\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpTaskwait();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp taskyield
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpTaskyield(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_taskyield) &&
                "Not a pragma omp \"taskyield\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpTaskyield();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp atomic
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpAtomic(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_atomic) &&
                "Not a pragma omp \"atomic\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpAtomic();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp flush ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpFlush(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_flush) &&
                "Not a pragma omp \"flush\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpFlush();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp ordered
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpOrdered(SourceLocation *TrailingElseLoc) {
  //FIXME
  assert(Tok.is(tok::pragma_omp_ordered) &&
                "Not a pragma omp \"ordered\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpOrdered();
}

//===----------------------------------------------------------------------===//
// Parse \#pragma omp threadprivate ...
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpThreadprivate(
                                         SourceLocation *TrailingElseLoc) {
  //FIXME: Many things
  assert(Tok.is(tok::pragma_omp_threadprivate) &&
                "Not a pragma omp \"threadprivate\" directive statement");
  Preprocessor &PP = getPreprocessor();
  PP.DiscardUntilEndOfDirective();
  ConsumeToken();
  return Actions.ActOnPragmaOmpThreadprivate();
}

//===----------------------------------------------------------------------===//
// Parsing methods for different OpenMP pragma clauses.
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause variable expression list.
// \Brief: This *overloaded* function assumes that left paranthesis is *not*
// consumed by the caller routine.
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpClauseVariableExpressions(ExprVector &DeclRefExprs,
                                      SourceLocation &LParenLoc,
                                      SourceLocation &RParenLoc) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause variable expression list.
// \Brief: This *overloaded* function assumes that left paranthesis is consumed
// by the caller routine.
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpClauseVariableExpressions(ExprVector &DeclRefExprs,
                                      SourceLocation &RParenLoc) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// \Brief: /{ OpenMP Clauses with a List of Variables. }/
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "private". 
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpPrivateClause(StmtVector &PrivateClauses) {
  //FIXME:Many things. 
  SourceLocation PrivateLoc = Tok.getLocation();

  ExprVector DeclRefExprs;
  SourceLocation LParenLoc, RParenLoc;
  ParsePragmaOmpClauseVariableExpressions(DeclRefExprs, LParenLoc, RParenLoc);

  StmtResult R = Actions.ActOnPragmaOmpClausePrivate(PrivateLoc, LParenLoc,
                                                     RParenLoc, DeclRefExprs);

  PrivateClauses.push_back(R.release());
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "firstprivate".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpFirstPrivateClause(StmtVector &FirstPrivateClauses) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "lastprivate".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpLastPrivateClause(StmtVector &LastPrivateClauses) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "shared".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpSharedClause(StmtVector &SharedClauses) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "copyin".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpCopyInClause(StmtVector &CopyInClauses) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "copyprivate".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpCopyPrivateClause(StmtVector &CopyPrivateClauses) {
  //FIXME
}

//===----------------------------------------------------------------------===//
// \Brief: /{ OpenMP Clauses with Scalar or Integer Expressions. }/
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "if".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpIfClause(StmtVector &IfClauses) {
  //FIXME : Many things.
  SourceLocation IfLoc = ConsumeToken();
  SourceLocation LParenLoc = NextToken().getLocation();

  //FIXME: Need to handle scope
  //Parse the condition.
  ExprResult CondExp;
  Decl *CondVar = 0;
  if (ParseParenExprOrCondition(CondExp, CondVar, IfLoc, true))
    return; 

  SourceLocation RParenLoc = CondExp.get()->getLocEnd();

  FullExprArg FullCondExp(Actions.MakeFullExpr(CondExp.get(), IfLoc));

  StmtResult R = Actions.ActOnPragmaOmpClauseIf(IfLoc, LParenLoc, RParenLoc,
                                                FullCondExp, CondVar);

  IfClauses.push_back(R.release());
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "num_threads".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpNumThreadsClause(StmtVector &NumThreadsClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "final".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpFinalClause(StmtVector &FinalClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "collapse".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpCollapseClause(StmtVector &CollapseClauses) {
}

//===----------------------------------------------------------------------===//
// \Brief: /{ OpenMP Compound Clauses. }/
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "reduction".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpReductionClause(StmtVector &ReductionClauses) {
  //FIXME
  SourceLocation RecuctionLoc = Tok.getLocation();
  SourceLocation LParenLoc = ConsumeToken();
  SourceLocation OperatorLoc = ConsumeToken();

  PragmaOmpHandler::ReductionOpKind OpKind;
  tok::TokenKind TokKind  = Tok.getKind();
  switch (TokKind) {
  case tok::plus:
    {
       OpKind = PragmaOmpHandler::R_PLUS;
       break;
    }
  case tok::minus:
    {
       OpKind = PragmaOmpHandler::R_MINUS;
       break;
    }
  case tok::star:
    {
       OpKind = PragmaOmpHandler::R_PROD;
       break;
    }
  default:
    {
       //Syntax error: Inavlid reduction operator.
    }
  }

  SourceLocation ColonLoc = ConsumeToken();

  ExprVector DeclRefExprs;
  SourceLocation RParenLoc;
  ParsePragmaOmpClauseVariableExpressions(DeclRefExprs, RParenLoc);

  StmtResult R = Actions.ActOnPragmaOmpClauseReduction(RecuctionLoc, LParenLoc,
                                                       OperatorLoc, ColonLoc,
                                                       RParenLoc, DeclRefExprs,
                                                       OpKind);

  ReductionClauses.push_back(R.release());

}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "schedule".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpScheduleClause(StmtVector &ScheduleClauses) {
}

//===----------------------------------------------------------------------===//
// \Brief: /{ OpenMP Clauses with Predefined Values. }/
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "default".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpDefaultClause(StmtVector &DefaultClauses) {
  //FIXME: Many things.

  // FIXME: If the vector DefaultClauses already holds a default clause AST
  // node, and if the control is again here, it means, "too many default
  // clauses". Report error, recover from error, and return from here.

  SourceLocation DefaultLoc = Tok.getLocation();
  SourceLocation LParenLoc = ConsumeToken();
  SourceLocation DefaultKindLoc = ConsumeToken();

  PragmaOmpHandler::DefaultKind Kind;
  StringRef Name = Tok.getIdentifierInfo()->getName();
  if (Omp.DefaultClause[PragmaOmpHandler::D_SHARED]->equals(Name)) {
    Kind = PragmaOmpHandler::D_SHARED;
  } else if (Omp.DefaultClause[PragmaOmpHandler::D_NONE]->equals(Name)) {
    Kind = PragmaOmpHandler::D_NONE;
  } else {
    // FIXME: Syntax error. Report error, recover from error, and return.
  }

  SourceLocation RParenLoc = ConsumeToken();

  StmtResult R = Actions.ActOnPragmaOmpClauseDefault(DefaultLoc, LParenLoc,
                                                     DefaultKindLoc, RParenLoc,
                                                     Kind);
  DefaultClauses.push_back(R.release());
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "ordered".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpOrderedClause(StmtVector &OrderedClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "nowait".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpNoWaitClause(StmtVector &NoWaitClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "untied".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpUntiedClause(StmtVector &UntiedClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "mergeable".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpMergeableClause(StmtVector &MergeableClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "read".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpReadClause(StmtVector &ReadClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "write".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpWriteClause(StmtVector &WriteClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "update".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpUpdateClause(StmtVector &UpdateClauses) {
}

//===----------------------------------------------------------------------===//
// Parse OpenMP pragma clause "capture".
//===----------------------------------------------------------------------===//
void Parser::ParsePragmaOmpCaptureClause(StmtVector &CaptureClauses) {
}

//===----------------------------------------------------------------------===//
// Parser::ParsePragmaOmpClause - \Brief: This is an entry point Parser routine
// to parse different OpenMP pragma clauses.
//===----------------------------------------------------------------------===//
StmtResult Parser::ParsePragmaOmpClauses() {
  // \Brief: AST nodes being created for defferent OpenMP clauses by Sema is
  // stored in an array of StmtVector. For the clauses, which can appear only
  // once in an OpenMP directive statement line, whose vector size is *one*. 
  // For other clauses, their vector sizes may greater than one depending on how
  // many times they appear on the OpenMP directive statement line. 
  StmtVector Clauses[PragmaOmpHandler::END_CLAUSE];

  // Record the start location of the first clause in the OpenMP
  // directive statement line. 
  SourceLocation StartLoc = NextToken().getLocation();
  SourceLocation EndLoc;
 
  // Parse all the OpenMP clauses present in the current OpenMP directive 
  // statement line.
  // Brief: Note that we won't introduce new *tokens* for openMP clause names
  // as these will get conflict with *identifier* token, and is very
  // complicated to handle, and not always neat. Instead, we parse (recognize)
  // the OpenMP clause names using string comparisons.
  // FIXME: Is it performance bottleneck? Do not think so.
  do {
    // Consume the next OpenMP clause name if any.
    ConsumeToken();
    tok::TokenKind Kind  = Tok.getKind();
    switch (Kind) {
    case tok::identifier:
      {
        // Parser has seen an identifier token. It must be one of the OpenMP
        // clause name. Parse it, if it is a valid OpenMP clause name.
        // Otherwise report a syntax error, and perform error recovery.
        StringRef Name = Tok.getIdentifierInfo()->getName();
        // OpenMP Clauses with a List of Variables.
        // Parse OpenMP pragma clause "private".
        if (Omp.Clause[PragmaOmpHandler::PRIVATE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpPrivateClause(
                     Clauses[PragmaOmpHandler::PRIVATE]);
        // Parse OpenMP pragma clause "firstprivate".
        } else if (Omp.Clause[PragmaOmpHandler::FIRST_PRIVATE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpFirstPrivateClause(
                     Clauses[PragmaOmpHandler::FIRST_PRIVATE]);
        // Parse OpenMP pragma clause "lastprivate".
        } else if (Omp.Clause[PragmaOmpHandler::LAST_PRIVATE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpLastPrivateClause(
                     Clauses[PragmaOmpHandler::LAST_PRIVATE]);
        // Parse OpenMP pragma clause "shared".
        } else if (Omp.Clause[PragmaOmpHandler::SHARED]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpSharedClause(
                     Clauses[PragmaOmpHandler::SHARED]);
        // Parse OpenMP pragma clause "copyin".
        } else if (Omp.Clause[PragmaOmpHandler::COPY_IN]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpCopyInClause(
                     Clauses[PragmaOmpHandler::COPY_IN]);
        // Parse OpenMP pragma clause "copyprivate".
        } else if (Omp.Clause[PragmaOmpHandler::COPY_PRIVATE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpCopyPrivateClause(
                     Clauses[PragmaOmpHandler::COPY_PRIVATE]);
        // OpenMP Clauses with Scalar or Integer Expressions.
        // Parse OpenMP pragma clause "if".
        } else if (Omp.Clause[PragmaOmpHandler::IF]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpIfClause(
                     Clauses[PragmaOmpHandler::IF]);
        // Parse OpenMP pragma clause "num_threads".
        } else if (Omp.Clause[PragmaOmpHandler::NUM_THREADS]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpNumThreadsClause(
                     Clauses[PragmaOmpHandler::NUM_THREADS]);
        // Parse OpenMP pragma clause "final".
        } else if (Omp.Clause[PragmaOmpHandler::FINAL]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpFinalClause(
                     Clauses[PragmaOmpHandler::FINAL]);
        // Parse OpenMP pragma clause "collapse".
        } else if (Omp.Clause[PragmaOmpHandler::COLLAPSE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpCollapseClause(
                     Clauses[PragmaOmpHandler::COLLAPSE]);
        // OpenMP Compound Clauses.
        // Parse OpenMP pragma clause "reduction".
        } else if (Omp.Clause[PragmaOmpHandler::REDUCTION]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpReductionClause(
                     Clauses[PragmaOmpHandler::REDUCTION]);
        // Parse OpenMP pragma clause "schedule".
        } else if (Omp.Clause[PragmaOmpHandler::SCHEDULE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpScheduleClause(
                     Clauses[PragmaOmpHandler::SCHEDULE]);
        // OpenMP Clauses with Predefined Values.
        // Parse OpenMP pragma clause "default".
        } else if (Omp.Clause[PragmaOmpHandler::DEFAULT]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpDefaultClause(
                     Clauses[PragmaOmpHandler::DEFAULT]);
        // Parse OpenMP pragma clause "ordered". 
        } else if (Omp.Clause[PragmaOmpHandler::ORDERED]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpOrderedClause(
                     Clauses[PragmaOmpHandler::ORDERED]);
        // Parse OpenMP pragma clause "nowait".
        } else if (Omp.Clause[PragmaOmpHandler::NOWAIT]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpNoWaitClause(
                     Clauses[PragmaOmpHandler::NOWAIT]);
        // Parse OpenMP pragma clause "untied".
        } else if (Omp.Clause[PragmaOmpHandler::UNTIED]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpUntiedClause(
                     Clauses[PragmaOmpHandler::UNTIED]);
        // Parse OpenMP pragma clause "mergeable".
        } else if (Omp.Clause[PragmaOmpHandler::MERGEABLE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpMergeableClause(
                     Clauses[PragmaOmpHandler::MERGEABLE]);
        // Parse OpenMP pragma clause "read".
        } else if (Omp.Clause[PragmaOmpHandler::READ]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpMergeableClause(
                     Clauses[PragmaOmpHandler::READ]);
        // Parse OpenMP pragma clause "write"
        } else if (Omp.Clause[PragmaOmpHandler::WRITE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpMergeableClause(
                     Clauses[PragmaOmpHandler::WRITE]);
        // Parse OpenMP pragma clause "update".
        } else if (Omp.Clause[PragmaOmpHandler::UPDATE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpMergeableClause(
                     Clauses[PragmaOmpHandler::UPDATE]);
        // Parse OpenMP pragma clause "capture".
        } else if (Omp.Clause[PragmaOmpHandler::CAPTURE]->equals(Name)) {
           // It could be last clause in the row. Record its location.
           EndLoc = Tok.getLocation();
           ParsePragmaOmpMergeableClause(
                     Clauses[PragmaOmpHandler::CAPTURE]);
        // Syntax error.
        } else {
          // Not a valid OpenMP pragma clause name. Report syntax error, and
          // recover from error.
          //FIXME
        }
        break;
      }
    case tok::eod:
      {
        // End of pragma omp directive line.
        // FIXME
        break;
      }
    default:
      {
        // Syntax error. Report a syntax error and perform error recovery.
        // FIXME
        break; 
      }
    }
  } while (Tok.isNot(tok::eod));

  // FIXME: How to pass an array of StmtVetor, and recieve it has an array
  // of MultiStmtArg. No implicit conversion possible?
  return Actions.ActOnPragmaOmpClauses(StartLoc, EndLoc,
                                       Clauses[PragmaOmpHandler::PRIVATE],
                                       Clauses[PragmaOmpHandler::FIRST_PRIVATE],
                                       Clauses[PragmaOmpHandler::LAST_PRIVATE],
                                       Clauses[PragmaOmpHandler::SHARED],
                                       Clauses[PragmaOmpHandler::COPY_IN],
                                       Clauses[PragmaOmpHandler::COPY_PRIVATE],
                                       Clauses[PragmaOmpHandler::IF],
                                       Clauses[PragmaOmpHandler::NUM_THREADS],
                                       Clauses[PragmaOmpHandler::FINAL],
                                       Clauses[PragmaOmpHandler::COLLAPSE],
                                       Clauses[PragmaOmpHandler::REDUCTION],
                                       Clauses[PragmaOmpHandler::SCHEDULE],
                                       Clauses[PragmaOmpHandler::DEFAULT],
                                       Clauses[PragmaOmpHandler::ORDERED],
                                       Clauses[PragmaOmpHandler::NOWAIT],
                                       Clauses[PragmaOmpHandler::UNTIED],
                                       Clauses[PragmaOmpHandler::MERGEABLE],
                                       Clauses[PragmaOmpHandler::READ],
                                       Clauses[PragmaOmpHandler::WRITE],
                                       Clauses[PragmaOmpHandler::UPDATE],
                                       Clauses[PragmaOmpHandler::CAPTURE]);
}
