//===----------------------- SemaStmtOpenMP.cpp ---------------------------===//
//          /{ Semantic Analysis for OpenMP Directive Statements }/
//
//                     The LLVM Compiler Infrastructure
//
// This file is distributed under the University of Illinois Open Source
// License. See LICENSE.TXT for details.
//
//===----------------------------------------------------------------------===//
//
//  This file implements semantic analysis for OpenMP directive statements.
//
//===----------------------------------------------------------------------===//

#include "clang/Sema/SemaInternal.h"
#include "clang/Sema/Scope.h"
#include "clang/Sema/ScopeInfo.h"
#include "clang/Sema/Initialization.h"
#include "clang/Sema/Lookup.h"
#include "clang/AST/ASTContext.h"
#include "clang/AST/CharUnits.h"
#include "clang/AST/DeclObjC.h"
#include "clang/AST/EvaluatedExprVisitor.h"
#include "clang/AST/ExprCXX.h"
#include "clang/AST/ExprObjC.h"
#include "clang/AST/StmtObjC.h"
#include "clang/AST/StmtCXX.h"
#include "clang/AST/StmtOpenMP.h"
#include "clang/AST/TypeLoc.h"
#include "clang/Lex/Preprocessor.h"
#include "clang/Basic/TargetInfo.h"
#include "llvm/ADT/ArrayRef.h"
#include "llvm/ADT/STLExtras.h"
#include "llvm/ADT/SmallPtrSet.h"
#include "llvm/ADT/SmallString.h"
#include "llvm/ADT/SmallVector.h"
using namespace clang;
using namespace sema;

//===----------------------------------------------------------------------===//
// Semantic analysis of OpenMP pragma directives.
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp parallel ... 
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpParallel(SourceLocation ParallelLoc, 
                                        Stmt* Clause, Stmt* ParallelBody) {
  // FIXME: All the semantics for OpenMP clauses, which are independent of
  // OpenMP directive statements are taken care while parsing clauses.
  // Here, *finish* semantics for OpenMP clauses, which are specific to current
  // OpenMP directive statement, that is /{ \#pragma omp parallel ... }/. 

  //FIXME: Other seamntic checks?

  //FIXME - Based on the changes within this routine.
  return Owned(new (Context) OmpParallelStmt(Context, Clause, ParallelBody,
                                             ParallelLoc));
}

//===----------------------------------------------------------------------===//
// Semantics fo \#pragam omp for ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpFor() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragam omp parallel for ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpParallelFor() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp sections ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpSections() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp section
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpSection() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp single ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpSingle() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp task ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpTask() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp master
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpMaster() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp critical ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpCritical() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp barrier
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpBarrier() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp taskwait
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpTaskwait() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp taskyield
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpTaskyield() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp atomic
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpAtomic() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp flush ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpFlush() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp ordered
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpOrdered() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantics for \#pragma omp threadprivate ...
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpThreadprivate() {
  //FIXME;
  return StmtEmpty();
}

//===----------------------------------------------------------------------===//
// Semantic analysis of OpenMP pragma clauses. 
//===----------------------------------------------------------------------===//

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "private".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClausePrivate(SourceLocation PrivateLoc,
                                             SourceLocation LParenLoc,
                                             SourceLocation RParenLoc, 
                                             MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpPrivateClause(Context, Exprs, NumExprs,
                       PrivateLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "firstprivate".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseFirstPrivate(
                                               SourceLocation FirstPrivateLoc,
                                               SourceLocation LParenLoc,
                                               SourceLocation RParenLoc,
                                               MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpFirstPrivateClause(Context, Exprs, NumExprs,
                       FirstPrivateLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "lastprivate".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseLastPrivate(SourceLocation LastPrivateLoc,
                                                 SourceLocation LParenLoc,
                                                 SourceLocation RParenLoc, 
                                                 MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpLastPrivateClause(Context, Exprs, NumExprs,
                       LastPrivateLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "shared".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseShared(SourceLocation SharedLoc,
                                            SourceLocation LParenLoc,
                                            SourceLocation RParenLoc,
                                            MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpSharedClause(Context, Exprs, NumExprs,
                       SharedLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===// 
// Semantics for OpenMP pragma clause "copyin".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseCopyIn(SourceLocation CopyInLoc,
                                            SourceLocation LParenLoc,
                                            SourceLocation RParenLoc,
                                            MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpCopyInClause(Context, Exprs, NumExprs,
                       CopyInLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "copyprivate".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseCopyPrivate(SourceLocation CopyPrivateLoc,
                                                 SourceLocation LParenLoc,
                                                 SourceLocation RParenLoc,
                                                 MultiExprArg DeclRefExprs) {
  //FIXME : Many things
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpCopyPrivateClause(Context, Exprs, NumExprs,
                       CopyPrivateLoc, LParenLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "if".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseIf(SourceLocation IfLoc,
                                        SourceLocation LParenLoc,
                                        SourceLocation RParenLoc,
                                        FullExprArg CondVal, Decl *CondVar) {
  //FIXME: Many things
  ExprResult CondResult(CondVal.release());

  VarDecl *ConditionVar = 0;
  if (CondVar) {
    ConditionVar = cast<VarDecl>(CondVar);
    CondResult = CheckConditionVariable(ConditionVar, IfLoc, true);
    if (CondResult.isInvalid())
      return StmtError();
  }
  Expr *ConditionExpr = CondResult.takeAs<Expr>();
  if (!ConditionExpr)
    return StmtError();


  return Owned(new (Context) OmpIfClause(Context, IfLoc, LParenLoc, RParenLoc,
                                         ConditionVar, ConditionExpr));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "reduction".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseReduction(SourceLocation RecuctionLoc,
                                     SourceLocation LParenLoc,
                                     SourceLocation OperatorLoc,
                                     SourceLocation ColonLoc,
                                     SourceLocation RParenLoc,
                                     MultiExprArg DeclRefExprs,
                                     PragmaOmpHandler::ReductionOpKind OpKind) {
  //FIXME
  unsigned NumExprs = DeclRefExprs.size();
  Expr **Exprs = DeclRefExprs.data();
  return Owned(new (Context) OmpReductionClause(Context, Exprs, NumExprs,
                      OpKind, RecuctionLoc, LParenLoc, OperatorLoc, ColonLoc,
                      RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause "default".
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauseDefault(SourceLocation DefaultLoc,
                                           SourceLocation LParenLoc,
                                           SourceLocation DefaultKindLoc,
                                           SourceLocation RParenLoc,
                                           PragmaOmpHandler::DefaultKind Kind) {
  //FIXME
  return Owned(new (Context) OmpDefaultClause(Context, Kind, DefaultLoc,
                       LParenLoc, DefaultKindLoc, RParenLoc));
}

//===----------------------------------------------------------------------===//
// Semantics for OpenMP pragma clause *list*.
//===----------------------------------------------------------------------===//
StmtResult Sema::ActOnPragmaOmpClauses(SourceLocation StartLoc,
                                       SourceLocation EndLoc,
                                       MultiStmtArg PrivateClauses,
                                       MultiStmtArg FirstPrivateClauses,
                                       MultiStmtArg LastPrivateClauses,
                                       MultiStmtArg SharedClauses, 
                                       MultiStmtArg CopyInClauses,
                                       MultiStmtArg CopyPrivateClauses, 
                                       MultiStmtArg IfClauses,
                                       MultiStmtArg NumThreadsClauses,
                                       MultiStmtArg FinalClauses,
                                       MultiStmtArg CollapseClauses, 
                                       MultiStmtArg ReductionClauses,
                                       MultiStmtArg ScheduleClauses, 
                                       MultiStmtArg DefaultClauses, 
                                       MultiStmtArg OrderedClauses,
                                       MultiStmtArg NoWaitClauses,
                                       MultiStmtArg UntiedClauses,
                                       MultiStmtArg MergeableClauses,
                                       MultiStmtArg ReadClauses,
                                       MultiStmtArg WriteClauses,
                                       MultiStmtArg UpdateClauses,
                                       MultiStmtArg CaptureClauses) {

  unsigned NumClauses[PragmaOmpHandler::END_CLAUSE];
  Stmt** Clauses[PragmaOmpHandler::END_CLAUSE];

  // Copy private clause vector. 
  if (PrivateClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::PRIVATE] = PrivateClauses.size();
    Clauses[PragmaOmpHandler::PRIVATE] = PrivateClauses.data(); 
  } else {
    NumClauses[PragmaOmpHandler::PRIVATE] = 0;
    Clauses[PragmaOmpHandler::PRIVATE] = 0;
  }

  // Copy firstprivate clause vector.
  if (FirstPrivateClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::FIRST_PRIVATE] = FirstPrivateClauses.size();
    Clauses[PragmaOmpHandler::FIRST_PRIVATE] = FirstPrivateClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::FIRST_PRIVATE] = 0;
    Clauses[PragmaOmpHandler::FIRST_PRIVATE] = 0;
  }

  // Copy lastprivate clause vector.
  if (LastPrivateClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::LAST_PRIVATE] = LastPrivateClauses.size();
    Clauses[PragmaOmpHandler::LAST_PRIVATE] = LastPrivateClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::LAST_PRIVATE] = 0;
    Clauses[PragmaOmpHandler::LAST_PRIVATE] = 0;
  }

  // Copy shared clause vector.
  if (SharedClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::SHARED] = SharedClauses.size();
    Clauses[PragmaOmpHandler::SHARED] = SharedClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::SHARED] = 0;
    Clauses[PragmaOmpHandler::SHARED] = 0;
  }

  // Copy copyin clause vector. 
  if (CopyInClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::COPY_IN] = CopyInClauses.size();
    Clauses[PragmaOmpHandler::COPY_IN] = CopyInClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::COPY_IN] = 0;
    Clauses[PragmaOmpHandler::COPY_IN] = 0;
  }

  // Copy copyprivate clause vector.
  if (CopyPrivateClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::COPY_PRIVATE] = CopyPrivateClauses.size();
    Clauses[PragmaOmpHandler::COPY_PRIVATE] = CopyPrivateClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::COPY_PRIVATE] = 0;
    Clauses[PragmaOmpHandler::COPY_PRIVATE] = 0;
  }

  // Copy if clause vector.
  if (IfClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::IF] = IfClauses.size();
    Clauses[PragmaOmpHandler::IF] = IfClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::IF] = 0;
    Clauses[PragmaOmpHandler::IF] = 0;
  }

  // Copy numthreads clause vector.
  if (NumThreadsClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::NUM_THREADS] = NumThreadsClauses.size();
    Clauses[PragmaOmpHandler::NUM_THREADS] = NumThreadsClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::NUM_THREADS] = 0;
    Clauses[PragmaOmpHandler::NUM_THREADS] = 0;
  }

  // Copy final clause vector.
  if (FinalClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::FINAL] = FinalClauses.size();
    Clauses[PragmaOmpHandler::FINAL] = FinalClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::FINAL] = 0;
    Clauses[PragmaOmpHandler::FINAL] = 0;
  }

  // Copy collapse clause vector.
  if (CollapseClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::COLLAPSE] = CollapseClauses.size();
    Clauses[PragmaOmpHandler::COLLAPSE] = CollapseClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::COLLAPSE] = 0;
    Clauses[PragmaOmpHandler::COLLAPSE] = 0;
  }

  // Copy reduction clause vector.
  if (ReductionClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::REDUCTION] = ReductionClauses.size();
    Clauses[PragmaOmpHandler::REDUCTION] = ReductionClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::REDUCTION] = 0;
    Clauses[PragmaOmpHandler::REDUCTION] = 0;
  }

  // Copy schedule clause vector.
  if (ScheduleClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::SCHEDULE] = ScheduleClauses.size();
    Clauses[PragmaOmpHandler::SCHEDULE] = ScheduleClauses.data();
  } else {
    NumClauses[PragmaOmpHandler::SCHEDULE] = 0;
    Clauses[PragmaOmpHandler::SCHEDULE] = 0;
  }

  // Copy default clause vector.
  if (DefaultClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::DEFAULT] = DefaultClauses.size();
    Clauses[PragmaOmpHandler::DEFAULT] = DefaultClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::DEFAULT] = 0;
    Clauses[PragmaOmpHandler::DEFAULT] = 0;
  }

  // Copy ordered clause vector.
  if (OrderedClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::ORDERED] = OrderedClauses.size();
    Clauses[PragmaOmpHandler::ORDERED] = OrderedClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::ORDERED] = 0;
    Clauses[PragmaOmpHandler::ORDERED] = 0;
  }

  // Copy nowait clause vector.
  if (NoWaitClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::NOWAIT] = NoWaitClauses.size();
    Clauses[PragmaOmpHandler::NOWAIT] = NoWaitClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::NOWAIT] = 0;
    Clauses[PragmaOmpHandler::NOWAIT] = 0;
  }

  // Copy untied clause vector.
  if (UntiedClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::UNTIED] = UntiedClauses.size();
    Clauses[PragmaOmpHandler::UNTIED] = UntiedClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::UNTIED] = 0;
    Clauses[PragmaOmpHandler::UNTIED] = 0;
  }

  // Copy mergeable clause vector.
  if (MergeableClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::MERGEABLE] = MergeableClauses.size();
    Clauses[PragmaOmpHandler::MERGEABLE] = MergeableClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::MERGEABLE] = 0;
    Clauses[PragmaOmpHandler::MERGEABLE] = 0;
  }

  // Copy read clause vector.
  if (ReadClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::READ] = ReadClauses.size();
    Clauses[PragmaOmpHandler::READ] = ReadClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::READ] = 0;
    Clauses[PragmaOmpHandler::READ] = 0;
  }

  // Copy write clause vector.
  if (WriteClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::WRITE] = WriteClauses.size();
    Clauses[PragmaOmpHandler::WRITE] = WriteClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::WRITE] = 0;
    Clauses[PragmaOmpHandler::WRITE] = 0;
  }

  // Copy update clause vector.
  if (UpdateClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::UPDATE] = UpdateClauses.size();
    Clauses[PragmaOmpHandler::UPDATE] = UpdateClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::UPDATE] = 0;
    Clauses[PragmaOmpHandler::UPDATE] = 0;
  }

  // Copy capture clause vector.
  if (CaptureClauses.size() > 0) {
    NumClauses[PragmaOmpHandler::CAPTURE] = CaptureClauses.size();
    Clauses[PragmaOmpHandler::CAPTURE] = CaptureClauses.data();

  } else {
    NumClauses[PragmaOmpHandler::CAPTURE] = 0;
    Clauses[PragmaOmpHandler::CAPTURE] = 0;
  }

  // FIXME: All the semantics for the clause list, which are independent of a
  // particular OpenMP directive statement are to be completed here. 

  return Owned(new (Context) OmpClause(Context, Clauses, NumClauses,
                    StartLoc, EndLoc));
}
